from requests.auth import AuthBase  # type: ignore


class BearerAuth(AuthBase):
    def __init__(self, access_token=None):
        self.access_token = access_token

    def __call__(self, req):
        if self.access_token:
            req.headers['Authorization'] = 'Bearer ' + self.access_token
        return req



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Timeout setting module

puppeteer equivalent: TimeoutSettings.js
"""

DEFAULT_TIMEOUT = 30_000  # 30 seconds


class TimeoutSettings:
    def __init__(self):
        self._defaultTimeout = None
        self._defaultNavigationTimeout = None

    def setDefaultTimeout(self, timeout: float) -> None:
        self._defaultTimeout = timeout

    def setDefaultNavigationTimeout(self, timeout: float) -> None:
        self._defaultNavigationTimeout = timeout

    @property
    def navigationTimeout(self) -> float:
        if self._defaultNavigationTimeout:
            return self._defaultNavigationTimeout
        if self._defaultTimeout:
            return self._defaultTimeout
        return DEFAULT_TIMEOUT

    @property
    def timeout(self) -> float:
        if self._defaultTimeout:
            return self._defaultTimeout
        return DEFAULT_TIMEOUT



import os
import shutil
import sys
import platform
import tempfile
import time
from pathlib import Path
from typing import Tuple
from zipfile import ZipFile
from termcolor import colored

import yaml

from biolib.biolib_api_client import AppVersionOnJob, Module, FilesMapping
from biolib.validators.validate_app_version import validate_app_version  # type: ignore
from biolib.validators.validate_argument import validate_argument  # type: ignore
from biolib.validators.validate_module import validate_module  # type: ignore
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger
from biolib.typing_utils import Dict, List


class BiolibValidationError(Exception):
    pass


def bl_colored(string, colour):
    # The colors does not seem to play nice with Windows PowerShell terminal
    if platform.system() == 'Windows':
        return string
    else:
        return colored(string, colour)


def validate_and_get_args(args):
    if args is None:
        args = []

    if isinstance(args, str):
        args = list(filter(lambda p: p != '', args.split(' ')))

    if not isinstance(args, list):
        raise Exception('The given input arguments must be list or str')

    return args


def get_files_dict_and_file_args(files, args):
    if files is None:
        files = []
        # TODO: Figure out how to make this slightly less error prone
        for idx, arg in enumerate(args):
            if not arg.startswith('/'):
                arg = f'{os.getcwd()}/{arg}'

            arg = Path(arg)
            if arg.is_file():
                files.append(arg)
                # Make sure that arg is only the filename
                args[idx] = arg.parts[-1]

    files_dict = {}
    for file_path in files:
        file = open(file_path, 'rb')
        path = '/' + file_path.parts[-1]

        files_dict[path] = file.read()
        file.close()

    return files_dict, args


def validate_and_get_app_path(provided_app_path):
    if provided_app_path.startswith('/'):
        app_path = Path(provided_app_path)
    else:
        app_path = Path(os.getcwd()) / provided_app_path

    if not app_path.exists():
        print(f'ERROR: Could not find path {app_path}')
        sys.exit(1)

    if not app_path.is_dir():
        print(f'ERROR: Path {app_path} is not a directory')
        sys.exit(1)

    return app_path


def get_yaml_data(app_path):
    try:
        yaml_file = open(f'{app_path}/.biolib/config.yml', 'r', encoding='utf-8')

    except Exception as error:  # pylint: disable=broad-except
        raise BioLibError(f'Could not open the biolib config file at {app_path}/.biolib/config.yml') from error

    try:
        yaml_data = yaml.safe_load(yaml_file)

    except Exception as error:  # pylint: disable=broad-except
        raise BioLibError(f'Could not parse {app_path}/.biolib/config.yml. Please make sure it is valid YAML') \
            from error

    return yaml_data


def validate_arguments(yaml_data):
    error_dict = {'arguments': {}}
    # Arguments are optional
    if 'arguments' in yaml_data:
        for argument in yaml_data['arguments']:
            argument_errors = validate_argument(argument)
            if argument_errors:
                error_dict['arguments'].update(argument_errors)

    # Just return empty dict if we have no errors
    if error_dict['arguments']:
        return error_dict
    else:
        return {}


def validate_and_get_biolib_yaml_version(yaml_data, error_dict):
    if not 'biolib_version' in yaml_data.keys():
        error_dict['config_yml']['biolib_version'] = ['Your config file is missing the biolib_version field.']
        print_errors(error_dict)
        sys.exit(1)
    else:
        biolib_version = yaml_data['biolib_version']

    if biolib_version not in (1, 2):
        error_dict['config_yml']['biolib_version'] = \
            [f'Biolib version can only be either 1 or 2 for now, your version is {biolib_version}.']
        print_errors(error_dict)
        sys.exit(1)

    return biolib_version


def validate_modules(yaml_data, biolib_yaml_version):
    error_dict = {'modules': {}}
    if 'modules' in yaml_data:
        for name, task_data in yaml_data['modules'].items():
            task_errors = validate_module(name, task_data, biolib_yaml_version)
            if task_errors:
                error_dict['modules'].update(task_errors)

    # Just return empty dict if we have no errors
    if error_dict['modules']:
        return error_dict
    else:
        return {}


def validate_yaml_config(yaml_data, source_files_zip_path):
    source_files_zip = ZipFile(source_files_zip_path)
    error_dict = {'config_yml': {}}

    biolib_yaml_version = validate_and_get_biolib_yaml_version(yaml_data, error_dict)
    error_dict['config_yml'].update(validate_app_version(yaml_data, biolib_yaml_version, source_files_zip))
    error_dict['config_yml'].update(validate_modules(yaml_data, biolib_yaml_version))
    error_dict['config_yml'].update(validate_arguments(yaml_data))

    if error_dict['config_yml']:
        print_errors(error_dict)
        raise BiolibValidationError('Could not validate .biolib/config.yml')


def print_errors(error_dict):
    print(bl_colored("\nThe following validation errors were found in the .biolib/config.yml file:", 'red'))
    print(yaml.safe_dump(error_dict, allow_unicode=True, default_flow_style=False))


def write_output_files(mapped_output_files=None):
    output_dir_path = 'biolib_results'

    if os.path.exists(output_dir_path):
        os.rename(output_dir_path, f'{output_dir_path}_old_{time.strftime("%Y%m%d%H%M%S")}')

    if mapped_output_files:
        logger.debug("Output Files:")

        for path, data in mapped_output_files.items():
            dirs, file = os.path.split(path)
            path_on_disk = f'{output_dir_path}{path}'

            if dirs:
                os.makedirs(f'{output_dir_path}{dirs}', exist_ok=True)
            if file:
                open(path_on_disk, 'wb').write(data)

            logger.debug(f"  - {path_on_disk}")

        logger.debug('\n')


def get_pretty_print_module_output_string(stdout=None, stderr=None, exitcode=None) -> str:
    stderr_string = f"\n{bl_colored('stderr:', 'red')}\n{stderr.decode()}" if stderr else ''
    return f'''{bl_colored('stdout:', 'green')}
{stdout.decode()}{stderr_string}
{bl_colored('exitcode:', 'green' if exitcode == 0 else 'red')} {exitcode}'''


def _get_files_mappings(file_commands: List[str]) -> List[FilesMapping]:
    files_mappings: List[FilesMapping] = []
    for file_command in file_commands:
        # TODO: figure out if splitting on space is good enough
        _, from_path, to_path = file_command.split(' ')
        files_mappings.append(FilesMapping(from_path=from_path, to_path=to_path))
    return files_mappings


def get_mocked_app_version_from_yaml(yaml_data: Dict, source_files_zip_path: str):
    modules: List[Module] = []
    for module_name in yaml_data['modules']:
        yaml_module = yaml_data['modules'][module_name]

        image_hostname, image_uri = yaml_module['image'].split('://')
        if image_hostname != 'local-docker':
            raise Exception(
                f'Found image {image_uri} with hostname {image_hostname}, please use "local-docker://"'
                f'Your application must use modules with local docker images to use run-dev'
            )

        modules.append(Module(
            command=yaml_module.get('command', ''),
            environment='biolib-ecr',
            image_uri=image_uri,
            input_files_mappings=_get_files_mappings(yaml_module['input_files']),
            large_file_systems=[],
            name=module_name,
            output_files_mappings=_get_files_mappings(yaml_module['output_files']),
            source_files_mappings=_get_files_mappings(yaml_module['source_files']),
            working_directory=yaml_module['working_directory'],
            estimated_image_size_bytes=None  # This should not be provided here
        ))

    return AppVersionOnJob(
        client_side_executable_zip=source_files_zip_path,
        consumes_stdin=True,
        is_runnable_by_user=True,
        modules=modules,
        public_id='app-version-mock-id',
        remote_hosts=[],
        settings=[],
        stdout_render_type='text',
    )


def create_temporary_directory_with_source_files_zip(app_path: str) -> Tuple[tempfile.TemporaryDirectory, str]:
    original_working_directory = os.getcwd()
    temporary_directory = tempfile.TemporaryDirectory()
    os.chdir(app_path)
    try:
        zip_path_without_file_extension = os.path.join(temporary_directory.name, 'biolib-cli-tmp-source-files-zip')
        zip_path = f'{zip_path_without_file_extension}.zip'
        app_folder_name = Path(app_path).absolute().parts[-1]
        shutil.make_archive(
            base_name=zip_path_without_file_extension,
            format='zip',
            root_dir='..',
            base_dir=app_folder_name,
        )

        return temporary_directory, zip_path

    except Exception as error:
        temporary_directory.cleanup()
        raise Exception('Failed to create zip of application') from error

    finally:
        # change back to old working directory
        os.chdir(original_working_directory)



import os
from json.decoder import JSONDecodeError

import requests
from biolib.typing_utils import Optional
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger


class _ApiClient:
    def __init__(self, base_url: str, access_token: Optional[str] = None):
        self.base_url: str = base_url
        self.refresh_token: Optional[str] = None
        self.access_token: Optional[str] = access_token

    @property
    def is_signed_in(self) -> bool:
        return self.refresh_token is not None and self.access_token is not None

    def login(self, api_token: Optional[str], exit_on_failure=False):
        if not api_token:
            if exit_on_failure:
                raise BioLibError('Error: Attempted login, but BIOLIB_TOKEN was not set, exiting...')
            else:
                logger.debug('Attempted login, but BIOLIB_TOKEN was not set, so continuing without logging in')
                return

        response = requests.post(
            f'{self.base_url}/api/user/api_tokens/exchange/',
            json={'token': api_token},
        )
        try:
            json_response = response.json()
        except JSONDecodeError as error:
            logger.error('Could not decode response from server as JSON:')
            raise BioLibError(response.text) from error
        if not response.ok:
            logger.error('Login with API token failed:')
            raise BioLibError(json_response['detail'])
        else:
            self.refresh_token = json_response['refresh_token']
            self.access_token = json_response['access_token']
            logger.info('Successfully authenticated')


class BiolibApiClient:
    api_client: Optional[_ApiClient] = None

    @staticmethod
    def initialize(base_url: str, access_token: Optional[str] = None):
        BiolibApiClient.api_client = _ApiClient(base_url, access_token)

    @staticmethod
    def get() -> _ApiClient:
        api_client = BiolibApiClient.api_client
        if api_client is not None:
            biolib_token = os.getenv('BIOLIB_TOKEN', default=None)
            if biolib_token is not None and not api_client.is_signed_in:
                api_client.login(api_token=biolib_token, exit_on_failure=True)
            return api_client
        else:
            raise BioLibError('Attempted to use uninitialized API client')

    @staticmethod
    def refresh_auth_token():
        if BiolibApiClient.api_client is None:
            raise BioLibError('Attempted to use uninitialized API client')

        BiolibApiClient.api_client.login(api_token=os.getenv('BIOLIB_TOKEN'), exit_on_failure=True)

    @staticmethod
    def assert_is_signed_in(authenticated_action_description: str) -> None:
        api_client = BiolibApiClient.get()
        if not api_client.is_signed_in:
            raise BioLibError(
                f'You must be signed in to {authenticated_action_description}. '
                f'Please set the environment variable "BIOLIB_TOKEN"'
            )



from biolib.biolib_binary_format.base_bbf_package import BioLibBinaryFormatBasePackage


class StdoutAndStderr(BioLibBinaryFormatBasePackage):
    def __init__(self, bbf=None):
        super().__init__(bbf)
        self.package_type = 10

    def serialize(self, stdout_and_stderr_bytes):
        bbf_data = bytearray()
        bbf_data.extend(self.version.to_bytes(1, 'big'))
        bbf_data.extend(self.package_type.to_bytes(1, 'big'))

        bbf_data.extend(len(stdout_and_stderr_bytes).to_bytes(8, 'big'))
        bbf_data.extend(stdout_and_stderr_bytes)

        return bbf_data

    def deserialize(self):
        version = self.get_data(1, output_type='int')
        package_type = self.get_data(1, output_type='int')
        self.check_version_and_type(version=version, package_type=package_type, expected_package_type=self.package_type)

        stdout_and_stderr_length = self.get_data(8, output_type='int')
        stdout_and_stderr = self.get_data(stdout_and_stderr_length)

        return stdout_and_stderr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Network Manager module."""

import asyncio
import base64
import json
import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Awaitable, Dict, List, Optional, Set, Union

from pyee import AsyncIOEventEmitter
from biolib.pyppeteer.pyppeteer.connection import CDPSession
from biolib.pyppeteer.pyppeteer.errors import NetworkError
from biolib.pyppeteer.pyppeteer.events import Events

if TYPE_CHECKING:
    from pyppeteer.frame import Frame, FrameManager

logger = logging.getLogger(__name__)


class NetworkManager(AsyncIOEventEmitter):
    """NetworkManager class."""

    def __init__(self, client: CDPSession, ignoreHttpsErrors: bool, frameManager: 'FrameManager') -> None:
        """Make new NetworkManager."""
        super().__init__()
        self._client = client
        self._ignoreHTTPSErrors = ignoreHttpsErrors
        self._frameManager = frameManager
        self._requestIdToRequest: Dict[Optional[str], Request] = {}
        self._requestIdToRequestWillBeSent: Dict[Optional[str], Dict] = {}
        self._extraHTTPHeaders: Dict[str, str] = {}
        self._offline: bool = False
        self._credentials: Optional[Dict[str, str]] = None
        self._attemptedAuthentications: Set[Optional[str]] = set()
        self._userRequestInterceptionEnabled = False
        self._protocolRequestInterceptionEnabled = False
        self._userCacheDisabled = False
        self._requestIdToInterceptionId: Dict[str, str] = {}

        self._client.on('Fetch.requestPaused', self._onRequestPaused)
        self._client.on('Fetch.authRequired', self._onAuthRequired)
        self._client.on('Network.requestWillBeSent', self._onRequestWillBeSent)
        self._client.on('Network.requestServedFromCache', self._onRequestServedFromCache)
        self._client.on('Network.responseReceived', self._onResponseReceived)
        self._client.on('Network.loadingFinished', self._onLoadingFinished)
        self._client.on('Network.loadingFailed', self._onLoadingFailed)

    async def initialize(self) -> None:
        await self._client.send('Network.enable')
        if self._ignoreHTTPSErrors:
            await self._client.send('Security.setIgnoreCertificateErrors', {'ignore': True})

    async def authenticate(self, credentials: Dict[str, str]) -> None:
        """Provide credentials for http auth."""
        self._credentials = credentials
        await self._updateProtocolRequestInterception()

    async def setExtraHTTPHeaders(self, extraHTTPHeaders: Dict[str, str]) -> None:
        """Set extra http headers."""
        self._extraHTTPHeaders = {}
        for k, v in extraHTTPHeaders.items():
            if not isinstance(v, str):
                raise TypeError(f'Expected value of header "{k}" to be string, but {type(v)} is found.')
            self._extraHTTPHeaders[k.lower()] = v
        await self._client.send('Network.setExtraHTTPHeaders', {'headers': self._extraHTTPHeaders})

    def extraHTTPHeaders(self) -> Dict[str, str]:
        """Get extra http headers."""
        return dict(**self._extraHTTPHeaders)

    async def setOfflineMode(self, value: bool) -> None:
        """Change offline mode enable/disable."""
        if self._offline == value:
            return
        self._offline = value
        await self._client.send(
            'Network.emulateNetworkConditions',
            {'offline': self._offline, 'latency': 0, 'downloadThroughput': -1, 'uploadThroughput': -1,},
        )

    async def setUserAgent(self, userAgent: str) -> None:
        """Set user agent."""
        await self._client.send('Network.setUserAgentOverride', {'userAgent': userAgent})

    async def setCacheEnabled(self, enabled: bool) -> None:
        self._userCacheDisabled = not enabled
        await self._updateProtocolCacheDisabled()

    async def setRequestInterception(self, value: bool) -> None:
        """Enable request interception."""
        self._userRequestInterceptionEnabled = value
        await self._updateProtocolRequestInterception()

    async def _updateProtocolRequestInterception(self) -> None:
        enabled = self._userRequestInterceptionEnabled or bool(self._credentials)
        if enabled == self._protocolRequestInterceptionEnabled:
            return
        self._protocolRequestInterceptionEnabled = enabled
        if enabled:
            await asyncio.gather(
                self._updateProtocolCacheDisabled(),
                self._client.send('Fetch.enable', {'handleAuthRequests': True, 'patterns': [{'urlPattern': '*'}]}),
            )
        else:
            await asyncio.gather(self._updateProtocolCacheDisabled(), self._client.send('Fetch.disable'))

    async def _updateProtocolCacheDisabled(self) -> None:
        await self._client.send(
            'Network.setCacheDisabled',
            {'cacheDisabled': self._userCacheDisabled or self._protocolRequestInterceptionEnabled},
        )

    async def _onRequestWillBeSent(self, event: Dict) -> None:
        is_data_request = event.get('request', {}).get('url', '').startswith('data:')
        if self._protocolRequestInterceptionEnabled and not is_data_request:
            requestId = event['requestId']
            interceptionId = self._requestIdToInterceptionId.get(requestId)
            if interceptionId:
                self._onRequest(event, interceptionId)
                self._requestIdToInterceptionId.pop(requestId)
            else:
                self._requestIdToRequestWillBeSent[requestId] = event
        else:
            self._onRequest(event, None)

    async def _onAuthRequired(self, event: Dict) -> None:
        response = 'Default'
        requestId = event.get('requestId')
        if requestId in self._attemptedAuthentications:
            response = 'CancelAuth'
        elif self._credentials:
            response = 'ProvideCredentials'
            self._attemptedAuthentications.add(requestId)
        if self._credentials:
            username = self._credentials.get('username')
            password = self._credentials.get('password')
        else:
            username, password = None, None
        await self._client.send(
            'Fetch.continueWithAuth',
            {
                'requestId': requestId,
                'authChallengeResponse': {"response": response, "username": username, "password": password,},
            },
        )

    async def _onRequestPaused(self, event: Dict) -> None:
        if not self._userRequestInterceptionEnabled and self._protocolRequestInterceptionEnabled:
            await self._client.send('Fetch.continueRequest', {'requestId': event.get('requestId')})
        requestId = event['networkId']
        interceptionId = event['requestId']
        if requestId in self._requestIdToRequestWillBeSent:
            requestWillBeSentEvent = self._requestIdToRequestWillBeSent.pop(requestId)
            self._onRequest(requestWillBeSentEvent, interceptionId)
        else:
            self._requestIdToInterceptionId[requestId] = interceptionId

    def _onRequest(self, event: Dict, interceptionId: Optional[str]) -> None:
        redirectChain: List[Request] = []
        if event.get('redirectResponse'):
            request = self._requestIdToRequest.get(event['requestId'])
            if request:
                self._handleRequestRedirect(request, event['redirectResponse'])
                redirectChain = request._redirectChain

        frame = self._frameManager.frame(event.get('frameId'))
        request = Request(
            client=self._client,
            frame=frame,
            interceptionId=interceptionId,
            allowInterception=self._userRequestInterceptionEnabled,
            event=event,
            redirectChain=redirectChain,
        )
        self._requestIdToRequest[event['requestId']] = request
        self.emit(Events.NetworkManager.Request, request)

    def _onRequestServedFromCache(self, event: Dict) -> None:
        request = self._requestIdToRequest.get(event.get('requestId'))
        if request:
            request._fromMemoryCache = True

    def _handleRequestRedirect(self, request: 'Request', responsePayload: Dict[str, Any]) -> None:
        response = Response(client=self._client, request=request, responsePayload=responsePayload)
        request._response = response
        request._redirectChain.append(request)
        response._bodyLoadedFutureFulFill(NetworkError('Response body is unavailable for redirect response'))
        self._requestIdToRequest.pop(request._requestId, None)
        self._attemptedAuthentications.discard(request._interceptionId)
        self.emit(Events.NetworkManager.Response, response)
        self.emit(Events.NetworkManager.RequestFinished, request)

    def _onResponseReceived(self, event: dict) -> None:
        request = self._requestIdToRequest.get(event['requestId'])
        # FileUpload sends a response without a matching request.
        if not request:
            return
        response = Response(self._client, request, event['response'])
        request._response = response
        self.emit(Events.NetworkManager.Response, response)

    def _onLoadingFinished(self, event: dict) -> None:
        request = self._requestIdToRequest.get(event['requestId'])
        # For certain requestIds we never receive requestWillBeSent event.
        # @see https://crbug.com/750469
        if not request:
            return
        response = request.response
        if response:
            response._bodyLoadedFutureFulFill(None)
        self._requestIdToRequest.pop(request._requestId, None)
        self._attemptedAuthentications.discard(request._interceptionId)
        self.emit(Events.NetworkManager.RequestFinished, request)

    def _onLoadingFailed(self, event: dict) -> None:
        request = self._requestIdToRequest.get(event['requestId'])
        # For certain requestIds we never receive requestWillBeSent event.
        # @see https://crbug.com/750469
        if not request:
            return
        request._failureText = event.get('errorText')
        response = request.response
        if response:
            response._bodyLoadedFutureFulFill(None)
        self._requestIdToRequest.pop(request._requestId, None)
        self._attemptedAuthentications.discard(request._interceptionId)
        self.emit(Events.NetworkManager.RequestFailed, request)


class Request:
    """Request class.

    Whenever the page sends a request, such as for a network resource, the
    following events are emitted by pyppeteer's page:

    - ``'request'``: emitted when the request is issued by the page.
    - ``'response'``: emitted when/if the response is received for the request.
    - ``'requestfinished'``: emitted when the response body is downloaded and
      the request is complete.

    If request fails at some point, then instead of ``'requestfinished'`` event
    (and possibly instead of ``'response'`` event), the ``'requestfailed'``
    event is emitted.

    If request gets a ``'redirect'`` response, the request is successfully
    finished with the ``'requestfinished'`` event, and a new request is issued
    to a redirect url.
    """

    _errorReasons = {
        'aborted': 'Aborted',
        'accessdenied': 'AccessDenied',
        'addressunreachable': 'AddressUnreachable',
        'blockedbyclient': 'BlockedByClient',
        'blockedbyresponse': 'BlockedByResponse',
        'connectionaborted': 'ConnectionAborted',
        'connectionclosed': 'ConnectionClosed',
        'connectionfailed': 'ConnectionFailed',
        'connectionrefused': 'ConnectionRefused',
        'connectionreset': 'ConnectionReset',
        'internetdisconnected': 'InternetDisconnected',
        'namenotresolved': 'NameNotResolved',
        'timedout': 'TimedOut',
        'failed': 'Failed',
    }

    def __init__(
        self,
        client: CDPSession,
        frame: Optional['Frame'],
        interceptionId: Optional[str],
        allowInterception: bool,
        event: Dict[str, Any],
        redirectChain: List['Request'],
    ):
        self._client = client
        self._requestId = event.get('requestId')
        self._isNavigationRequest = self._requestId == event.get('loaderId') and event['type'] == 'Document'
        self._interceptionId = interceptionId
        self._allowInterception = allowInterception
        self._interceptionHandled = False
        self._response = None
        self._failureText = None

        request_event = event['request']
        self._url = request_event['url']
        self._resourceType = event.get('type').lower() if isinstance(event.get('type'), str) else None
        self._method = request_event['method']
        self._postData = request_event.get('postData')
        self._frame = frame
        self._redirectChain = redirectChain
        self._headers = {k.lower(): v for k, v in request_event.get('headers', {}).items()}

        self._fromMemoryCache = False

    @property
    def url(self) -> str:
        return self._url

    @property
    def resourceType(self) -> str:
        """Resource type of this request perceived by the rendering engine.

        ResourceType will be one of the following: ``document``,
        ``stylesheet``, ``image``, ``media``, ``font``, ``script``,
        ``texttrack``, ``xhr``, ``fetch``, ``eventsource``, ``websocket``,
        ``manifest``, ``other``.
        """
        return self._resourceType

    @property
    def method(self) -> Optional[str]:
        """Return this request's method (GET, POST, etc.)."""
        return self._method

    @property
    def postData(self) -> Optional[str]:
        """Return post body of this request."""
        return self._postData

    @property
    def headers(self) -> Dict[str, str]:
        """Return a dictionary of HTTP headers of this request.

        All header names are lower-case.
        """
        return self._headers

    @property
    def response(self) -> Optional['Response']:
        """Return matching :class:`Response` object, or ``None``.

        If the response has not been received, return ``None``.
        """
        return self._response

    @property
    def frame(self) -> Optional['Frame']:
        """Return a matching :class:`~pyppeteer.frame_manager.frame` object.

        Return ``None`` if navigating to error page.
        """
        return self._frame

    @property
    def isNavigationRequest(self) -> bool:
        """Whether this request is driving frame's navigation."""
        return self._isNavigationRequest

    @property
    def redirectChain(self) -> List['Request']:
        """Return chain of requests initiated to fetch a resource.

        * If there are no redirects and request was successful, the chain will
          be empty.
        * If a server responds with at least a single redirect, then the chain
          will contain all the requests that were redirected.

        ``redirectChain`` is shared between all the requests of the same chain.
        """
        return self._redirectChain.copy()

    @property
    def failure(self) -> Optional[Dict[str, str]]:
        """Return error text.

        Return ``None`` unless this request was failed, as reported by
        ``requestfailed`` event.

        When request failed, this method return dictionary which has a
        ``errorText`` field, which contains human-readable error message, e.g.
        ``'net::ERR_RAILED'``.
        """
        return None if not self._failureText else {'errorText': self._failureText}

    async def continue_(
        self, url: str = None, method: str = None, postData: str = None, headers: Dict[str, str] = None
    ) -> None:
        """Continue request with optional request overrides.

        To use this method, request interception should be enabled by
        :meth:`pyppeteer.page.Page.setRequestInterception`. If request
        interception is not enabled, raise ``NetworkError``.

        ``overrides`` can have the following fields:

        * ``url`` (str): If set, the request url will be changed.
        * ``method`` (str): If set, change the request method (e.g. ``GET``).
        * ``postData`` (str): If set, change the post data or request.
        * ``headers`` (dict): If set, change the request HTTP header.
        """
        if self._is_actionable_request:
            if headers:
                headers = headersArray(headers)

            self._interceptionHandled = True
            try:
                await self._client.send(
                    'Fetch.continueRequest',
                    {
                        'requestId': self._interceptionId,
                        'url': url,
                        'method': method,
                        'postData': postData,
                        'headers': headers,
                    },
                )
            except Exception as e:
                # In certain cases, protocol will return error if the request was already canceled
                # or the page was closed. We should tolerate these errors.
                # (logger.exception will pick up the stack trace for us)
                logger.exception(f'An exception occurred while trying to continue the request')

    async def respond(
        self, status: int = 200, headers: Dict[str, str] = None, contentType: str = None, body: Union[bytes, str] = None
    ) -> None:
        """Fulfills request with given response.

        To use this, request interception should by enabled by
        :meth:`pyppeteer.page.Page.setRequestInterception`. Request
        interception is not enabled, raise ``NetworkError``.

        * ``status`` (int): Response status code, defaults to 200.
        * ``headers`` (dict): Optional response headers.
        * ``contentType`` (str): If set, equals to setting ``Content-Type``
          response header.
        * ``body`` (str|bytes): Optional response body.
        """
        if self._is_actionable_request:
            self._interceptionHandled = True

            if isinstance(body, str):
                responseBody = body.encode('utf-8')
            else:
                responseBody = body

            responseHeaders = {k.lower(): v for k, v in (headers or {}).items()}
            if contentType:
                responseHeaders['content-type'] = contentType
            if responseBody and 'content-length' not in responseHeaders:
                responseHeaders['content-length'] = str(len(responseBody))
            try:
                await self._client.send(
                    'Fetch.fulfillRequest',
                    {
                        'requestId': self._interceptionId,
                        'responseCode': status,
                        'responsePhrase': STATUS_TEXTS[status],
                        'responseHeaders': headersArray(responseHeaders),
                        'body': base64.b64encode(responseBody).decode('ascii') if responseBody else None,
                    },
                )
            except Exception as e:
                # In certain cases, protocol will return error if the request was already canceled
                # or the page was closed. We should tolerate these errors.
                logger.error(f'An exception occurred fulfilling a response: {e}')

    async def abort(self, errorCode: str = 'failed') -> None:
        """Abort request.

        To use this, request interception should be enabled by
        :meth:`pyppeteer.page.Page.setRequestInterception`.
        If request interception is not enabled, raise ``NetworkError``.

        ``errorCode`` is an optional error code string. Defaults to ``failed``,
        could be one of the following:

        - ``aborted``: An operation was aborted (due to user action).
        - ``accessdenied``: Permission to access a resource, other than the
          network, was denied.
        - ``addressunreachable``: The IP address is unreachable. This usually
          means that there is no route to the specified host or network.
        - ``blockedbyclient``: The client chose to block the request.
        - ``blockedbyresponse``: The request failed because the request was
          delivered along with requirements which are not met
          ('X-Frame-Options' and 'Content-Security-Policy' ancestor check,
          for instance).
        - ``connectionaborted``: A connection timeout as a result of not
          receiving an ACK for data sent.
        - ``connectionclosed``: A connection was closed (corresponding to a TCP
          FIN).
        - ``connectionfailed``: A connection attempt failed.
        - ``connectionrefused``: A connection attempt was refused.
        - ``connectionreset``: A connection was reset (corresponding to a TCP
          RST).
        - ``internetdisconnected``: The Internet connection has been lost.
        - ``namenotresolved``: The host name could not be resolved.
        - ``timedout``: An operation timed out.
        - ``failed``: A generic failure occurred.
        """
        if self._is_actionable_request:
            errorReason = self._errorReasons.get(errorCode)
            if not errorReason:
                raise NetworkError(f'Unknown error code: {errorCode}')

            self._interceptionHandled = True
            try:
                await self._client.send(
                    'Fetch.failRequest', {'requestId': self._interceptionId, 'errorReason': errorReason}
                )
            except Exception as e:
                # In certain cases, protocol will return error if the request was already canceled
                # or the page was closed. We should tolerate these errors.
                logger.error(f'An exception occurred: {e}')

    @property
    def _is_actionable_request(self) -> bool:
        """
        Checks if we can abort/continue/respond to request.
        :return: True if we can, False if we can't
        """
        # Mocking responses for dataURL requests is not currently supported.
        if self._url.startswith('data:'):
            return False
        if not self._allowInterception:
            raise ValueError('Request Interception is not enabled!')
        if self._interceptionHandled:
            raise ValueError('Request is already handled')
        return True


class Response:
    """Response class represents responses which are received by ``Page``."""

    def __init__(self, client: CDPSession, request: 'Request', responsePayload):
        self._client = client
        self._request = request
        self._contentFuture = None

        self._bodyLoadedFuture = client.loop.create_future()
        self._bodyLoadedFutureFulFill = lambda x: self._bodyLoadedFuture.set_result(x)

        self._remoteAddress = {
            'ip': responsePayload.get('remoteIPAddress'),
            'port': responsePayload.get('remotePort'),
        }
        self._status = responsePayload.get('status')
        self._statusText = responsePayload.get('statusText')
        self._url = request.url
        self._fromDiskCache = bool(responsePayload.get('fromDiskCache'))
        self._fromServiceWorker = bool(responsePayload.get('fromServiceWorker'))
        self._headers = {k.lower(): v for k, v in responsePayload.get('headers', {}).items()}
        if responsePayload.get('securityDetails'):
            self._securityDetails = SecurityDetails(**responsePayload.get('securityDetails'))
        else:
            self._securityDetails = None

    @property
    def remoteAddress(self) -> Dict[str, str]:
        return self._remoteAddress

    @property
    def url(self) -> str:
        return self._url

    @property
    def ok(self) -> bool:
        return self._status == 0 or (200 <= self._status < 300)

    @property
    def status(self) -> int:
        return self._status

    @property
    def statusText(self) -> str:
        return self._statusText

    @property
    def headers(self) -> Dict[str, str]:
        """
        Return dictionary of HTTP headers of this response.
        All header names are lower-case.
        """
        return self._headers

    @property
    def securityDetails(self) -> Optional['SecurityDetails']:
        """Return security details associated with this response.

        Security details if the response was received over the secure
        connection, or `None` otherwise.
        """
        return self._securityDetails

    def buffer(self) -> Awaitable[bytes]:
        """Return awaitable which resolves to bytes with response body."""
        if self._contentFuture is None:

            async def buffer_read():
                await self._bodyLoadedFuture
                response = await self._client.send('Network.getResponseBody', {'requestId': self._request._requestId})
                body = response.get('body', '')
                if response.get('base64Encoded'):
                    return base64.b64decode(body)
                # b64decode gives us bytes, we encode this so that this fn always returns a bytes
                return body.encode('utf-8')

            self._contentFuture = self._client.loop.create_task(buffer_read())
            return self._contentFuture

    @property
    async def text(self) -> str:
        """Text representation of response body."""
        return (await self.buffer()).decode('utf-8', errors='replace')

    @property
    async def json(self) -> Dict[str, Any]:
        """JSON representation of response body."""
        return json.loads(await self.text)

    @property
    def request(self) -> 'Request':
        """matching :class:`Request` object."""
        return self._request

    @property
    def fromCache(self) -> bool:
        """
        Return ``True`` if the response was served from cache.
        Here `cache` is either the browser's disk cache or memory cache.
        """
        return self._fromDiskCache or self._request._fromMemoryCache

    @property
    def fromServiceWorker(self) -> bool:
        """Return ``True`` if the response was served by a service worker."""
        return self._fromServiceWorker

    @property
    def frame(self) -> Optional['Frame']:
        return self._request.frame


class SecurityDetails:
    """Class represents responses which are received by page."""

    _recorded_attributes = {'subjectName', 'issuer', 'validFrom', 'validTo', 'protocol'}

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if k in self._recorded_attributes:
                self.__setattr__(f'_{k}', v)

    @property
    def subjectName(self) -> str:
        """Return the subject to which the certificate was issued to."""
        return self._subjectName

    @property
    def issuer(self) -> str:
        """Return a string with the name of issuer of the certificate."""
        return self._issuer

    @property
    def validFrom(self) -> int:
        """Return UnixTime of the start of validity of the certificate."""
        return self._validFrom

    @property
    def validTo(self) -> int:
        """Return UnixTime of the end of validity of the certificate."""
        return self._validTo

    @property
    def protocol(self) -> str:
        """Return string of with the security protocol, e.g. "TLS1.2"."""
        return self._protocol


def headersArray(headers: Dict[str, str]) -> List[Dict[str, str]]:
    return [{'name': k, 'value': str(v)} for k, v in headers.items() if v is not None]


STATUS_TEXTS = {num.value: code for code, num in vars(HTTPStatus).items() if isinstance(num, HTTPStatus)}
